@@ -1,5 +1,5 @@
 // Symphonia
-// Copyright (c) 2020 The Project Symphonia Developers.
+// Copyright (c) 2019-2021 The Project Symphonia Developers.
 //
 // Previous Author: Kostya Shishkov <kostya.shiskov@gmail.com>
 //
@@ -15,17 +15,18 @@ use std::f32::consts;
 use std::fmt;
 
 use symphonia_core::errors::{decode_error, unsupported_error, Result};
-use symphonia_core::io::{huffman::*, BitStream, FiniteBitStream, BitStreamLtr, BufStream};
+use symphonia_core::io::{ReadBitsLtr, FiniteBitStream, BitReaderLtr};
+use symphonia_core::io::vlc::{Codebook, Entry16x16};
 use symphonia_core::audio::{AudioBuffer, AudioBufferRef, AsAudioBufferRef, Signal, SignalSpec};
-use symphonia_core::codecs::CODEC_TYPE_AAC;
-use symphonia_core::codecs::{CodecParameters, CodecDescriptor, Decoder, DecoderOptions};
+use symphonia_core::codecs::{CODEC_TYPE_AAC, CodecParameters, CodecDescriptor};
+use symphonia_core::codecs::{Decoder, DecoderOptions, FinalizeResult};
 use symphonia_core::dsp::mdct::Imdct;
 use symphonia_core::formats::Packet;
 use symphonia_core::support_codec;
 use symphonia_core::units::Duration;
 
+use super::codebooks;
 use super::common::*;
-use super::huffman_tables::*;
 use super::window::*;
 
 use lazy_static::lazy_static;
@@ -35,31 +36,17 @@ macro_rules! validate {
     ($a:expr) => {
         if !$a {
             error!("check failed at {}:{}", file!(), line!());
-            return decode_error("invalid data");
+            return decode_error("aac: invalid data");
         }
     };
 }
 
-const SPEC_TABLES: [&'static HuffmanTable<H16>; 11] = [
-    &SPEC_TABLE_1,
-    &SPEC_TABLE_2,
-    &SPEC_TABLE_3,
-    &SPEC_TABLE_4,
-    &SPEC_TABLE_5,
-    &SPEC_TABLE_6,
-    &SPEC_TABLE_7,
-    &SPEC_TABLE_8,
-    &SPEC_TABLE_9,
-    &SPEC_TABLE_10,
-    &SPEC_TABLE_11,
-];
-
 lazy_static! {
     /// Pre-computed table of y = x^(4/3).
     static ref POW43_TABLE: [f32; 8192] = {
         let mut pow43 = [0f32; 8192];
-        for i in 0..8192 {
-            pow43[i] = f32::powf(i as f32, 4.0 / 3.0);
+        for (i, pow43) in pow43.iter_mut().enumerate() {
+            *pow43 = f32::powf(i as f32, 4.0 / 3.0);
         }
         pow43
     };
@@ -95,7 +82,7 @@ impl M4AInfo {
         }
     }
 
-    fn read_object_type<B: BitStream>(bs: &mut B) -> Result<M4AType> {
+    fn read_object_type<B: ReadBitsLtr>(bs: &mut B) -> Result<M4AType> {
         let otypeidx = match bs.read_bits_leq32(5)? {
             idx if idx < 31 => idx as usize,
             31 => (bs.read_bits_leq32(6)? + 32) as usize,
@@ -109,7 +96,7 @@ impl M4AInfo {
         }
     }
 
-    fn read_sampling_frequency<B: BitStream>(bs: &mut B) -> Result<u32> {
+    fn read_sampling_frequency<B: ReadBitsLtr>(bs: &mut B) -> Result<u32> {
         match bs.read_bits_leq32(4)? {
             idx if idx < 15 => Ok(AAC_SAMPLE_RATES[idx as usize]),
             _ => {
@@ -119,7 +106,7 @@ impl M4AInfo {
         }
     }
 
-    fn read_channel_config<B: BitStream>(bs: &mut B) -> Result<usize> {
+    fn read_channel_config<B: ReadBitsLtr>(bs: &mut B) -> Result<usize> {
         let chidx = bs.read_bits_leq32(4)? as usize;
         if chidx < AAC_CHANNELS.len() {
             Ok(AAC_CHANNELS[chidx])
@@ -130,7 +117,7 @@ impl M4AInfo {
     }
 
     fn read(&mut self, buf: &[u8]) -> Result<()> {
-        let mut bs = BitStreamLtr::new(BufStream::new(buf));
+        let mut bs = BitReaderLtr::new(buf);
 
         self.otype = Self::read_object_type(&mut bs)?;
         self.srate = Self::read_sampling_frequency(&mut bs)?;
@@ -139,7 +126,7 @@ impl M4AInfo {
 
         self.channels = Self::read_channel_config(&mut bs)?;
 
-        if (self.otype == M4AType::SBR) || (self.otype == M4AType::PS) {
+        if (self.otype == M4AType::Sbr) || (self.otype == M4AType::PS) {
             let ext_srate = Self::read_sampling_frequency(&mut bs)?;
             self.otype = Self::read_object_type(&mut bs)?;
 
@@ -155,8 +142,8 @@ impl M4AInfo {
 
         match self.otype {
             M4AType::Main
-            | M4AType::LC
-            | M4AType::SSR
+            | M4AType::Lc
+            | M4AType::Ssr
             | M4AType::Scalable
             | M4AType::TwinVQ
             | M4AType::ER_AAC_LC
@@ -179,7 +166,7 @@ impl M4AInfo {
                 let extension_flag = bs.read_bit()?;
 
                 if self.channels == 0 {
-                    return unsupported_error("program config element");
+                    return unsupported_error("aac: program config element");
                 }
 
                 if (self.otype == M4AType::Scalable) || (self.otype == M4AType::ER_AAC_Scalable) {
@@ -205,59 +192,59 @@ impl M4AInfo {
                     let extension_flag3 = bs.read_bit()?;
 
                     if extension_flag3 {
-                        return unsupported_error("version3 extensions");
+                        return unsupported_error("aac: version3 extensions");
                     }
                 }
             }
-            M4AType::CELP => {
-                return unsupported_error("CELP config");
+            M4AType::Celp => {
+                return unsupported_error("aac: CELP config");
             }
-            M4AType::HVXC => {
-                return unsupported_error("HVXC config");
+            M4AType::Hvxc => {
+                return unsupported_error("aac: HVXC config");
             }
-            M4AType::TTSI => {
-                return unsupported_error("TTS config");
+            M4AType::Ttsi => {
+                return unsupported_error("aac: TTS config");
             }
             M4AType::MainSynth
             | M4AType::WavetableSynth
             | M4AType::GeneralMIDI
             | M4AType::Algorithmic => {
-                return unsupported_error("structured audio config");
+                return unsupported_error("aac: structured audio config");
             }
             M4AType::ER_CELP => {
-                return unsupported_error("ER CELP config");
+                return unsupported_error("aac: ER CELP config");
             }
             M4AType::ER_HVXC => {
-                return unsupported_error("ER HVXC config");
+                return unsupported_error("aac: ER HVXC config");
             }
             M4AType::ER_HILN | M4AType::ER_Parametric => {
-                return unsupported_error("parametric config");
+                return unsupported_error("aac: parametric config");
             }
-            M4AType::SSC => {
-                return unsupported_error("SSC config");
+            M4AType::Ssc => {
+                return unsupported_error("aac: SSC config");
             }
             M4AType::MPEGSurround => {
                 // bs.ignore_bits(1)?; // sacPayloadEmbedding
-                return unsupported_error("MPEG Surround config");
+                return unsupported_error("aac: MPEG Surround config");
             }
             M4AType::Layer1 | M4AType::Layer2 | M4AType::Layer3 => {
-                return unsupported_error("MPEG Layer 1/2/3 config");
+                return unsupported_error("aac: MPEG Layer 1/2/3 config");
             }
-            M4AType::DST => {
-                return unsupported_error("DST config");
+            M4AType::Dst => {
+                return unsupported_error("aac: DST config");
             }
-            M4AType::ALS => {
+            M4AType::Als => {
                 // bs.ignore_bits(5)?; // fillBits
-                return unsupported_error("ALS config");
+                return unsupported_error("aac: ALS config");
             }
-            M4AType::SLS | M4AType::SLSNonCore => {
-                return unsupported_error("SLS config");
+            M4AType::Sls | M4AType::SLSNonCore => {
+                return unsupported_error("aac: SLS config");
             }
             M4AType::ER_AAC_ELD => {
-                return unsupported_error("ELD config");
+                return unsupported_error("aac: ELD config");
             }
             M4AType::SMRSimple | M4AType::SMRMain => {
-                return unsupported_error("symbolic music config");
+                return unsupported_error("aac: symbolic music config");
             }
             _ => {}
         };
@@ -277,7 +264,7 @@ impl M4AInfo {
                 let ep_config = bs.read_bits_leq32(2)?;
 
                 if (ep_config == 2) || (ep_config == 3) {
-                    return unsupported_error("error protection config");
+                    return unsupported_error("aac: error protection config");
                 }
                 // if ep_config == 3 {
                 //     let direct_mapping = bs.read_bit()?;
@@ -292,7 +279,7 @@ impl M4AInfo {
 
             if sync == 0x2B7 {
                 let ext_otype = Self::read_object_type(&mut bs)?;
-                if ext_otype == M4AType::SBR {
+                if ext_otype == M4AType::Sbr {
                     self.sbr_present = bs.read_bit()?;
                     if self.sbr_present {
                         let _ext_srate = Self::read_sampling_frequency(&mut bs)?;
@@ -367,12 +354,13 @@ impl ICSInfo {
             long_win: true,
         }
     }
-    fn decode_ics_info<B: BitStream>(&mut self, bs: &mut B) -> Result<()> {
+
+    fn decode_ics_info<B: ReadBitsLtr>(&mut self, bs: &mut B) -> Result<()> {
         self.prev_window_sequence = self.window_sequence;
         self.prev_window_shape = self.window_shape;
 
         if bs.read_bit()? {
-            return decode_error("ics reserved bit set");
+            return decode_error("aac: ics reserved bit set");
         }
 
         self.window_sequence = bs.read_bits_leq32(2)? as u8;
@@ -441,12 +429,13 @@ impl ICSInfo {
 struct LTPData {}
 
 impl LTPData {
-    fn read<B: BitStream>(bs: &mut B) -> Result<Option<Self>> {
+    fn read<B: ReadBitsLtr>(bs: &mut B) -> Result<Option<Self>> {
         let predictor_data_present = bs.read_bit()?;
         if !predictor_data_present {
             return Ok(None);
         }
-        return unsupported_error("predictor data");
+
+        unsupported_error("aac: predictor data")
         /*
                 if is_main {
                     let predictor_reset                         = bs.read_bit()?;
@@ -484,7 +473,7 @@ struct PulseData {
 }
 
 impl PulseData {
-    fn read<B: BitStream>(bs: &mut B) -> Result<Option<Self>> {
+    fn read<B: ReadBitsLtr>(bs: &mut B) -> Result<Option<Self>> {
         let pulse_data_present = bs.read_bit()?;
         if !pulse_data_present {
             return Ok(None);
@@ -530,7 +519,7 @@ impl TNSCoeffs {
             coef: [0.0; TNS_MAX_ORDER + 1],
         }
     }
-    fn read<B: BitStream>(
+    fn read<B: ReadBitsLtr>(
         &mut self,
         bs: &mut B,
         long_win: bool,
@@ -579,9 +568,7 @@ impl TNSCoeffs {
                 for i in 1..m {
                     b[i] = self.coef[i - 1] + tmp[m - 1] * self.coef[m - i - 1];
                 }
-                for i in 1..m {
-                    self.coef[i - 1] = b[i];
-                }
+                self.coef[..(m - 1)].copy_from_slice(&b[1..m]);
                 self.coef[m - 1] = tmp[m - 1];
             }
         }
@@ -600,7 +587,7 @@ struct TNSData {
 }
 
 impl TNSData {
-    fn read<B: BitStream>(
+    fn read<B: ReadBitsLtr>(
         bs: &mut B,
         long_win: bool,
         num_windows: usize,
@@ -637,12 +624,12 @@ struct GainControlData {
 }
 
 impl GainControlData {
-    fn read<B: BitStream>(bs: &mut B) -> Result<Option<Self>> {
+    fn read<B: ReadBitsLtr>(bs: &mut B) -> Result<Option<Self>> {
         let gain_control_data_present = bs.read_bit()?;
         if !gain_control_data_present {
             return Ok(None);
         }
-        return unsupported_error("gain control data");
+        unsupported_error("aac: gain control data")
         /*        self.max_band                                   = bs.read_bits_leq32(2)? as u8;
                 if window_sequence == ONLY_LONG_SEQUENCE {
                     for bd in 0..max_band
@@ -661,7 +648,7 @@ const INTENSITY_HCB2: u8 = 14;
 const INTENSITY_HCB: u8 = 15;
 
 #[derive(Clone)]
-struct ICS {
+struct Ics {
     global_gain: u8,
     info: ICSInfo,
     pulse_data: Option<PulseData>,
@@ -692,7 +679,7 @@ fn get_intensity_scale(scale: i16) -> f32 {
     2.0f32.powf(0.25 * f32::from(scale))
 }
 
-impl ICS {
+impl Ics {
     fn new(sbinfo: GASubbandInfo) -> Self {
         Self {
             global_gain: 0,
@@ -712,7 +699,12 @@ impl ICS {
         }
     }
 
-    fn decode_section_data<B: BitStream>(
+    fn reset(&mut self) {
+        self.info = ICSInfo::new();
+        self.delay = [0.0; 1024];
+    }
+
+    fn decode_section_data<B: ReadBitsLtr>(
         &mut self,
         bs: &mut B
     ) -> Result<()> {
@@ -728,7 +720,7 @@ impl ICS {
                 self.sect_len[g][l] = 0;
 
                 if self.sect_cb[g][l] == RESERVED_HCB {
-                    return decode_error("invalid band type");
+                    return decode_error("aac: invalid band type");
                 }
 
                 loop {
@@ -776,12 +768,14 @@ impl ICS {
         self.sfb_cb[g][sfb] == INTENSITY_HCB
     }
 
-    fn decode_scale_factor_data<B: BitStream>(&mut self, bs: &mut B) -> Result<()> {
+    fn decode_scale_factor_data<B: ReadBitsLtr>(&mut self, bs: &mut B) -> Result<()> {
         let mut noise_pcm_flag = true;
         let mut scf_intensity = 0i16;
         let mut scf_noise = i16::from(self.global_gain) - 90;
         let mut scf_normal = i16::from(self.global_gain);
 
+        let scf_table = &codebooks::SCF_CODEBOOK;
+
         for g in 0..self.info.window_groups {
             for sfb in 0..self.info.max_sfb {
 
@@ -789,8 +783,7 @@ impl ICS {
                     0.0
                 }
                 else if self.is_intensity(g, sfb) {
-                    // TODO: Fix lim_bits.
-                    scf_intensity += i16::from(bs.read_huffman(&SCF_TABLE, 100)?.0) - 60;
+                    scf_intensity += i16::from(bs.read_codebook(scf_table)?.0) - 60;
 
                     validate!(
                         (scf_intensity >= INTENSITY_SCALE_MIN)
@@ -805,8 +798,7 @@ impl ICS {
                         scf_noise += (bs.read_bits_leq32(9)? as i16) - 256;
                     }
                     else {
-                        // TODO: Fix lim_bits.
-                        scf_noise += i16::from(bs.read_huffman(&SCF_TABLE, 100)?.0) - 60;
+                        scf_noise += i16::from(bs.read_codebook(scf_table)?.0) - 60;
                     }
 
                     validate!(
@@ -816,8 +808,7 @@ impl ICS {
                     get_scale(scf_noise)
                 }
                 else {
-                    // TODO: Fix lim_bits.
-                    scf_normal += i16::from(bs.read_huffman(&SCF_TABLE, 100)?.0) - 60;
+                    scf_normal += i16::from(bs.read_codebook(scf_table)?.0) - 60;
                     validate!((scf_normal >= 0) && (scf_normal < 255));
 
                     get_scale(scf_normal - 100)
@@ -846,7 +837,7 @@ impl ICS {
         }
     }
 
-    fn decode_spectrum<B: BitStream>(&mut self, bs: &mut B) -> Result<()> {
+    fn decode_spectrum<B: ReadBitsLtr>(&mut self, bs: &mut B) -> Result<()> {
         // Zero all spectral coefficients.
         self.coeffs = [0.0; 1024];
         for g in 0..self.info.window_groups {
@@ -868,7 +859,7 @@ impl ICS {
                         _ => {
                             let unsigned = AAC_UNSIGNED_CODEBOOK[(cb_idx - 1) as usize];
 
-                            let cb = SPEC_TABLES[(cb_idx - 1) as usize];
+                            let cb = &codebooks::SPECTRUM_CODEBOOKS[(cb_idx - 1) as usize];
 
                             if cb_idx < FIRST_PAIR_HCB {
                                 decode_quads(bs, cb, unsigned, scale, dst)?;
@@ -931,7 +922,7 @@ impl ICS {
         }
     }
 
-    fn decode_ics<B: BitStream>(
+    fn decode_ics<B: ReadBitsLtr>(
         &mut self,
         bs: &mut B,
         m4atype: M4AType,
@@ -954,7 +945,7 @@ impl ICS {
         let tns_max_order = if !self.info.long_win {
             7
         }
-        else if m4atype == M4AType::LC {
+        else if m4atype == M4AType::Lc {
             12
         }
         else {
@@ -965,7 +956,7 @@ impl ICS {
             TNSData::read(bs, self.info.long_win, self.info.num_windows, tns_max_order)?;
 
         match m4atype {
-            M4AType::SSR => self.gain_control = GainControlData::read(bs)?,
+            M4AType::Ssr => self.gain_control = GainControlData::read(bs)?,
             _ => {
                 let gain_control_data_present = bs.read_bit()?;
                 validate!(!gain_control_data_present);
@@ -976,7 +967,7 @@ impl ICS {
         Ok(())
     }
 
-    fn synth_channel(&mut self, dsp: &mut DSP, srate_idx: usize, dst: &mut [f32]) {
+    fn synth_channel(&mut self, dsp: &mut Dsp, srate_idx: usize, dst: &mut [f32]) {
         self.place_pulses();
 
         if let Some(ref tns_data) = self.tns_data {
@@ -1094,13 +1085,13 @@ fn decode_spectrum_noise(lcg: &mut Lcg, sf: f32, dst: &mut [f32]) {
     let scale = sf / energy.sqrt();
 
     for spec in dst.iter_mut() {
-        *spec = *spec * scale;
+        *spec *= scale;
     }
 }
 
-fn decode_quads<B: BitStream>(
+fn decode_quads<B: ReadBitsLtr>(
     bs: &mut B,
-    cb: &'static HuffmanTable<H16>,
+    cb: &Codebook<Entry16x16>,
     unsigned: bool,
     scale: f32,
     dst: &mut [f32],
@@ -1109,31 +1100,28 @@ fn decode_quads<B: BitStream>(
     let pow43_table: &[f32; 8192] = &POW43_TABLE;
 
     for out in dst.chunks_mut(4) {
-        // TODO: Fix lim_bits
-        let cw = bs.read_huffman(cb, 100)?.0 as usize;
+        let cw = bs.read_codebook(cb)?.0 as usize;
         if unsigned {
-            for i in 0..4 {
-                let val = AAC_QUADS[cw][i];
-
-                if val != 0 {
-                    if bs.read_bit()? {
-                        out[i] = -pow43_table[val as usize] * scale;
+            for (out, &quad) in out.iter_mut().zip(&AAC_QUADS[cw]) {
+                if quad != 0 {
+                    *out = if bs.read_bit()? {
+                        -pow43_table[quad as usize] * scale
                     }
                     else {
-                        out[i] = pow43_table[val as usize] * scale;
+                        pow43_table[quad as usize] * scale
                     }
                 }
             }
         }
         else {
-            for i in 0..4 {
-                let val = AAC_QUADS[cw][i] - 1;
+            for (out, &quad) in out.iter_mut().zip(&AAC_QUADS[cw]) {
+                let val = quad - 1;
 
-                if val < 0 {
-                    out[i] = -pow43_table[-val as usize] * scale;
+                *out = if val < 0 {
+                    -pow43_table[-val as usize] * scale
                 }
                 else {
-                    out[i] = pow43_table[val as usize] * scale;
+                    pow43_table[val as usize] * scale
                 }
             }
         }
@@ -1141,9 +1129,9 @@ fn decode_quads<B: BitStream>(
     Ok(())
 }
 
-fn decode_pairs<B: BitStream>(
+fn decode_pairs<B: ReadBitsLtr>(
     bs: &mut B,
-    cb: &'static HuffmanTable<H16>,
+    cb: &Codebook<Entry16x16>,
     unsigned: bool,
     escape: bool,
     modulo: u16,
@@ -1154,8 +1142,7 @@ fn decode_pairs<B: BitStream>(
     let pow43_table: &[f32; 8192] = &POW43_TABLE;
 
     for out in dst.chunks_mut(2) {
-        // TODO: Fix lim_bits
-        let cw = bs.read_huffman(cb, 100)?.0;
+        let cw = bs.read_codebook(cb)?.0;
 
         let mut x = (cw / modulo) as i16;
         let mut y = (cw % modulo) as i16;
@@ -1188,12 +1175,8 @@ fn decode_pairs<B: BitStream>(
     Ok(())
 }
 
-fn read_escape<B: BitStream>(bs: &mut B, is_pos: bool) -> Result<i16> {
-    // TODO: Move into BitReaderLtr
-    let mut n = 0;
-    while bs.read_bit()? == true {
-        n += 1;
-    }
+fn read_escape<B: ReadBitsLtr>(bs: &mut B, is_pos: bool) -> Result<i16> {
+    let n = bs.read_unary_ones()?;
 
     validate!(n < 9);
 
@@ -1215,8 +1198,8 @@ struct ChannelPair {
     common_window: bool,
     ms_mask_present: u8,
     ms_used: [[bool; MAX_SFBS]; MAX_WINDOWS],
-    ics0: ICS,
-    ics1: ICS,
+    ics0: Ics,
+    ics1: Ics,
 }
 
 impl ChannelPair {
@@ -1227,17 +1210,22 @@ impl ChannelPair {
             common_window: false,
             ms_mask_present: 0,
             ms_used: [[false; MAX_SFBS]; MAX_WINDOWS],
-            ics0: ICS::new(sbinfo),
-            ics1: ICS::new(sbinfo),
+            ics0: Ics::new(sbinfo),
+            ics1: Ics::new(sbinfo),
         }
     }
 
-    fn decode_ga_sce<B: BitStream>(&mut self, bs: &mut B, m4atype: M4AType) -> Result<()> {
+    fn reset(&mut self) {
+        self.ics0.reset();
+        self.ics1.reset();
+    }
+
+    fn decode_ga_sce<B: ReadBitsLtr>(&mut self, bs: &mut B, m4atype: M4AType) -> Result<()> {
         self.ics0.decode_ics(bs, m4atype, false)?;
         Ok(())
     }
 
-    fn decode_ga_cpe<B: BitStream>(&mut self, bs: &mut B, m4atype: M4AType) -> Result<()> {
+    fn decode_ga_cpe<B: ReadBitsLtr>(&mut self, bs: &mut B, m4atype: M4AType) -> Result<()> {
         let common_window = bs.read_bit()?;
         self.common_window = common_window;
 
@@ -1263,7 +1251,7 @@ impl ChannelPair {
                         }
                     }
                 }
-                3 => return decode_error("invalid mid-side mask"),
+                3 => return decode_error("aac: invalid mid-side mask"),
                 _ => unreachable!()
             }
 
@@ -1332,7 +1320,7 @@ impl ChannelPair {
         Ok(())
     }
 
-    fn synth_audio(&mut self, dsp: &mut DSP, abuf: &mut AudioBuffer<f32>, srate_idx: usize) {
+    fn synth_audio(&mut self, dsp: &mut Dsp, abuf: &mut AudioBuffer<f32>, srate_idx: usize) {
         self.ics0.synth_channel(dsp, srate_idx, abuf.chan_mut(self.channel));
 
         if self.is_pair {
@@ -1341,7 +1329,7 @@ impl ChannelPair {
     }
 }
 
-struct DSP {
+struct Dsp {
     kbd_long_win: [f32; 1024],
     kbd_short_win: [f32; 128],
     sine_long_win: [f32; 1024],
@@ -1355,7 +1343,7 @@ struct DSP {
 const SHORT_WIN_POINT0: usize = 512 - 64;
 const SHORT_WIN_POINT1: usize = 512 + 64;
 
-impl DSP {
+impl Dsp {
     fn new() -> Self {
         let mut kbd_long_win: [f32; 1024] = [0.0; 1024];
         let mut kbd_short_win: [f32; 128] = [0.0; 128];
@@ -1390,7 +1378,7 @@ impl DSP {
         }
     }
 
-    #[allow(clippy::cyclomatic_complexity)]
+    #[allow(clippy::cognitive_complexity)]
     fn synth(
         &mut self,
         coeffs: &[f32; 1024],
@@ -1449,18 +1437,15 @@ impl DSP {
                 }
             }
             EIGHT_SHORT_SEQUENCE => {
-                for i in 0..SHORT_WIN_POINT0 {
-                    dst[i] = delay[i];
-                }
+                dst[..SHORT_WIN_POINT0].copy_from_slice(&delay[..SHORT_WIN_POINT0]);
 
                 for i in SHORT_WIN_POINT0..1024 {
                     dst[i] = delay[i] + self.ew_buf[i - SHORT_WIN_POINT0];
                 }
             }
             LONG_STOP_SEQUENCE => {
-                for i in 0..SHORT_WIN_POINT0 {
-                    dst[i] = delay[i];
-                }
+                dst[..SHORT_WIN_POINT0].copy_from_slice(&delay[..SHORT_WIN_POINT0]);
+
                 for i in SHORT_WIN_POINT0..SHORT_WIN_POINT1 {
                     dst[i] = delay[i] + self.tmp[i] * prev_short_win[i - SHORT_WIN_POINT0];
                 }
@@ -1488,9 +1473,8 @@ impl DSP {
                 }
             }
             LONG_START_SEQUENCE => {
-                for i in 0..SHORT_WIN_POINT0 {
-                    delay[i] = self.tmp[i + 1024];
-                }
+                delay[..SHORT_WIN_POINT0].copy_from_slice(&self.tmp[1024..(SHORT_WIN_POINT0 + 1024)]);
+
                 for i in SHORT_WIN_POINT0..SHORT_WIN_POINT1 {
                     delay[i] = self.tmp[i + 1024] * short_win[127 - (i - SHORT_WIN_POINT0)];
                 }
@@ -1511,7 +1495,7 @@ pub struct AacDecoder {
     // info: NACodecInfoRef,
     m4ainfo: M4AInfo,
     pairs: Vec<ChannelPair>,
-    dsp: DSP,
+    dsp: Dsp,
     sbinfo: GASubbandInfo,
     params: CodecParameters,
     buf: AudioBuffer<f32>,
@@ -1533,7 +1517,7 @@ impl AacDecoder {
         Ok(())
     }
 
-    fn decode_ga<B: BitStream + FiniteBitStream>(&mut self, bs: &mut B) -> Result<()> {
+    fn decode_ga<B: ReadBitsLtr + FiniteBitStream>(&mut self, bs: &mut B) -> Result<()> {
         let mut cur_pair = 0;
         let mut cur_ch = 0;
         while bs.bits_left() > 3 {
@@ -1558,7 +1542,7 @@ impl AacDecoder {
                 }
                 2 => {
                     // ID_CCE
-                    return unsupported_error("coupling channel element");
+                    return unsupported_error("aac: coupling channel element");
                 }
                 3 => {
                     // ID_LFE
@@ -1583,7 +1567,7 @@ impl AacDecoder {
                 }
                 5 => {
                     // ID_PCE
-                    return unsupported_error("program config");
+                    return unsupported_error("aac: program config");
                 }
                 6 => {
                     // ID_FIL
@@ -1634,7 +1618,7 @@ impl Decoder for AacDecoder {
         else {
             // Otherwise, assume there is no ASC and use the codec parameters for ADTS.
             m4ainfo.srate = params.sample_rate.unwrap();
-            m4ainfo.otype = M4AType::LC;
+            m4ainfo.otype = M4AType::Lc;
             m4ainfo.samples = 1024;
             m4ainfo.channels = params.channels.unwrap().count();
         }
@@ -1643,8 +1627,8 @@ impl Decoder for AacDecoder {
 
         trace!("{}", m4ainfo);
 
-        if (m4ainfo.otype != M4AType::LC) || (m4ainfo.channels > 2) || (m4ainfo.samples != 1024) {
-            return unsupported_error("aac too complex");
+        if (m4ainfo.otype != M4AType::Lc) || (m4ainfo.channels > 2) || (m4ainfo.samples != 1024) {
+            return unsupported_error("aac: aac too complex");
         }
 
         let spec = SignalSpec::new(
@@ -1658,13 +1642,19 @@ impl Decoder for AacDecoder {
         Ok(AacDecoder {
             m4ainfo,
             pairs: Vec::new(),
-            dsp: DSP::new(),
+            dsp: Dsp::new(),
             sbinfo: GASubbandInfo::find(srate),
             params: params.clone(),
             buf: AudioBuffer::new(duration, spec),
         })
     }
 
+    fn reset(&mut self) {
+        for pair in self.pairs.iter_mut() {
+            pair.reset();
+        }
+    }
+
     fn supported_codecs() -> &'static [CodecDescriptor] {
         &[
             support_codec!(CODEC_TYPE_AAC, "aac", "Advanced Audio Coding"),
@@ -1676,25 +1666,23 @@ impl Decoder for AacDecoder {
     }
 
     fn decode(&mut self, packet: &Packet) -> Result<AudioBufferRef<'_>> {
-        let reader = packet.as_buf_stream();
-
         // Clear the audio output buffer.
         self.buf.clear();
         self.buf.render_reserved(None);
 
-        let mut bs = BitStreamLtr::new(reader);
+        let mut bs = BitReaderLtr::new(packet.buf());
 
         // Choose decode step based on the object type.
         match self.m4ainfo.otype {
-            M4AType::LC => self.decode_ga(&mut bs)?,
-            _           => return unsupported_error("object type"),
+            M4AType::Lc => self.decode_ga(&mut bs)?,
+            _           => return unsupported_error("aac: object type"),
         }
 
         Ok(self.buf.as_audio_buffer_ref())
     }
 
-    fn close(&mut self) {
-
+    fn finalize(&mut self) -> FinalizeResult {
+        Default::default()
     }
 }
 
